""" Timing related generator functions """

from dune.codegen.generation import (cached,
                                     include_file,
                                     pre_include,
                                     preamble,
                                     )
from dune.codegen.options import get_option
from dune.codegen.pdelab.driver import (get_form_ident,
                                        is_linear,
                                        name_initree,
                                        name_mpihelper,
                                        )
from dune.codegen.pdelab.driver.gridfunctionspace import (name_leafview,
                                                          name_trial_gfs,
                                                          type_leafview,
                                                          )
from dune.codegen.pdelab.driver.gridoperator import (name_gridoperator,
                                                     name_localoperator,
                                                     type_gridoperator,
                                                     )
from dune.codegen.pdelab.driver.solve import (name_vector,
                                              define_vector,
                                              )


@preamble(section="timings")
def define_timing_identifier(name):
    ini = name_initree()
    return "auto {} = {}.get<std::string>(\"identifier\", std::string(argv[0]));".format(name, ini)


def name_timing_identifier():
    name = "ident"
    define_timing_identifier(name)
    return name


@preamble(section="timings")
def dump_dof_numbers(stream):
    ident = name_timing_identifier()
    level = get_option("instrumentation_level")
    include_file("dune/pdelab/common/partitionviewentityset.hh", filetag="driver")
    gvt = type_leafview()
    gv = name_leafview()

    from dune.codegen.pdelab.driver import get_trial_element, isDG, _flatten_list
    from ufl import MixedElement, TensorProductElement
    element = get_trial_element()

    def _apply_to_element(element, f):
        if isinstance(element, MixedElement) or isinstance(element, TensorProductElement):
            return tuple(_apply_to_element(e, f) for e in element.sub_elements())
        else:
            return f(element)
    element_is_dg = all(_flatten_list(_apply_to_element(element, isDG)))

    if element_is_dg:
        return ["Dune::PDELab::NonOverlappingEntitySet<{}> es({});".format(gvt, gv),
                "{} << \"{} \" << {} << \" dofs dofs \" << {}.maxLocalSize() * es.size(0) << std::endl;".format(stream,
                                                                                                                level,
                                                                                                                ident,
                                                                                                                name_trial_gfs())
                ]
    else:
        return ["{} << \"{} \" << {} << \" dofs dofs \" << {}.size() << std::endl;".format(stream, level, ident,
                                                                                           name_trial_gfs())
                ]


@preamble(section="timings")
def define_timing_stream(name):
    include_file('fstream', filetag='driver', system=True)
    include_file('sstream', filetag='driver', system=True)
    include_file('sys/types.h', filetag='driver', system=True)
    include_file('chrono', filetag='driver', system=True)

    return ["std::stringstream ss;",
            "ss << \"{}/timings-rank-\" << {}.rank() << \"-\" << std::chrono::high_resolution_clock::now().time_since_epoch().count() << \".csv\";".format(get_option('project_basedir'), name_mpihelper()),
            "std::ofstream {};".format(name),
            "{}.open(ss.str(), std::ios_base::app);".format(name),
            ]


def name_timing_stream():
    name = "timestream"
    define_timing_stream(name)
    dump_dof_numbers(name)
    return name


def name_temporary_vector(name, form):
    name = "{}_{}".format(name, form)
    define_vector(name, form)
    return name


@preamble(section="timings")
def define_jacobian(name, form_ident):
    t_go = type_gridoperator(form_ident)
    n_go = name_gridoperator(form_ident)
    return ["using M_{} = typename {}::Traits::Jacobian;".format(form_ident, t_go),
            "M_{} {}({});".format(form_ident, name, n_go)]


def name_jacobian(form_ident):
    name = "J_{}".format(form_ident)
    define_jacobian(name, form_ident)
    return name


@preamble(section="init")
def init_likwid():
    return ["LIKWID_MARKER_INIT;", "LIKWID_MARKER_THREADINIT;"]


@preamble(section="end")
def finalize_likwid():
    return ["LIKWID_MARKER_CLOSE;"]


@cached
def driver_likwid():
    @preamble(section="init")
    def start_driver_likwid():
        return ["LIKWID_MARKER_START(\"driver\");"]

    @preamble(section="end")
    def stop_driver_likwid():
        return ["LIKWID_MARKER_STOP(\"driver\");"]

    start_driver_likwid()
    stop_driver_likwid()


@cached
def setup_timer():
    # TODO check that we are using YASP?
    if get_option("use_likwid"):
        pre_include("#define LIKWID_PERFMON", filetag="driver")
        include_file("likwid.h", filetag="driver")
        init_likwid()
        driver_likwid()
        finalize_likwid()
    else:
        from dune.codegen.loopy.target import type_floatingpoint
        pre_include("#define HP_TIMER_OPCOUNTER {}".format(type_floatingpoint()), filetag="driver")
        if get_option('opcounter'):
            pre_include("#define ENABLE_COUNTER", filetag="driver")
        pre_include("#define ENABLE_HP_TIMERS", filetag="driver")
        include_file("dune/codegen/common/timer.hh", filetag="driver")



@preamble(section="init")
def init_likwid_timer(region):
    return ["LIKWID_MARKER_REGISTER(\"{}\");".format(region)]


def init_region_timer(region):
    setup_timer()
    if get_option("use_likwid"):
        init_likwid_timer(region)
    else:
        from dune.codegen.generation import post_include
        post_include("HP_DECLARE_TIMER({});".format(region), filetag="driver")


def start_region_timer(region):
    if get_option("use_likwid"):
        return ["LIKWID_MARKER_START(\"{}\");".format(region)]
    else:
        return ["HP_TIMER_START({});".format(region)]


def stop_region_timer(region):
    if get_option("use_likwid"):
        return ["LIKWID_MARKER_STOP(\"{}\");".format(region)]
    else:
        timestream = name_timing_stream()
        return ["HP_TIMER_STOP({});".format(region),
                "DUMP_TIMER({}, {}, {}, true);".format(get_option("instrumentation_level"), region, timestream)]


def timed_region(region, actions):
    if isinstance(actions, str):
        actions = [actions]

    assert(isinstance(actions, list))

    if get_option('instrumentation_level') >= 2:
        assembly = []
        print_times = []

        init_region_timer(region)

        if get_option('instrumentation_level') >= 3:
            timestream = name_timing_stream()
            lop_name = name_localoperator(get_form_ident())
            print_times.append("{}.dump_timers({}, {}, true);".format(lop_name, timestream, name_timing_identifier()))

        assembly += start_region_timer(region)
        assembly += actions
        assembly += stop_region_timer(region)

        return assembly + print_times
    else:
        return actions


@preamble(section="timings")
def evaluate_residual_timer():
    n_go = name_gridoperator(get_form_ident())
    v = name_vector(get_form_ident())
    r = name_temporary_vector("r", get_form_ident())

    action = "{}.residual({}, {});".format(n_go, v, r)

    return timed_region("residual_evaluation", action)


@preamble(section="timings")
def apply_jacobian_timer():
    form = get_form_ident()
    n_go = name_gridoperator(form)
    v = name_vector(form)

    if is_linear():
        j = name_temporary_vector("j", form)
        action = "{}.jacobian_apply({}, {});".format(n_go, v, j)
    else:
        j0 = name_temporary_vector("j0", form)
        j1 = name_temporary_vector("j1", form)
        action = "{}.nonlinear_jacobian_apply({}, {}, {});".format(n_go, v, j0, j1)

    return timed_region("apply_jacobian", action)


@preamble(section="timings")
def assemble_matrix_timer():
    n_go = name_gridoperator(get_form_ident())
    v = name_vector(get_form_ident())
    m = name_jacobian(get_form_ident())

    action = "{}.jacobian({},{});".format(n_go, v, m)

    return timed_region("matrix_assembly", action)
